#include "gui.h"

#include <cstdio>
#include <cstdlib>
#include <fstream>
#include <sys/stat.h>
#include <unistd.h>
#include <QFileInfo>

GryfApp *app;
QTextStream debug(stdout);

// GryfApp

GryfApp::GryfApp(int argc, char **argv) : QApplication(argc, argv)
{
	pgm = new program;

	cell_i = cell_j = -1;
	dir_i = 1, dir_j = 0;

	cell_grid = new QGridLayout();
	cell_grid->setHorizontalSpacing(0);
	cell_grid->setVerticalSpacing(0);

	palette_idle.setColor(QPalette::Background, Qt::white);
	palette_idle.setColor(QPalette::Foreground, Qt::black);

	palette_active.setColor(QPalette::Background, QColor(128, 0, 64));
	palette_active.setColor(QPalette::Foreground, Qt::white);

	for (int i=0; i < NUM_ROWS; i++) {
		for (int j=0; j < NUM_COLS; j++) {
			GryfCell *cell = new GryfCell;
			cell->app = this;
			cell->i = i;
			cell->j = j;
			cell->setFrameShape(QFrame::Box);
			cell->setPalette(palette_idle);
			cell->setAutoFillBackground(true);
			cell->setTextFormat(Qt::RichText);
			cell->render();
			cells[i][j] = cell;
			cell_grid->addWidget(cell, i, j, 0);
		}
	}

	status_line = new QLabel;
	status_line->setSizePolicy(QSizePolicy::Policy::Preferred, QSizePolicy::Policy::Fixed);
	status("Gryf v-1.-1 ready.");

	grid_widget = new GryfGrid;
	grid_widget->app = this;
	grid_widget->setLayout(cell_grid);
	grid_widget->setFocusPolicy(Qt::StrongFocus);

	left_vbox = new QVBoxLayout;
	left_vbox->addWidget(grid_widget);
	left_vbox->addWidget(status_line);

	right_vbox = new QVBoxLayout;

	QLabel *label_team = new QLabel("T&eam:");
	right_vbox->addWidget(label_team);

	edit_team = new QLineEdit();
	edit_team->setMaxLength(4);
	right_vbox->addWidget(edit_team);
	label_team->setBuddy(edit_team);
	QObject::connect(edit_team, SIGNAL(returnPressed()), this, SLOT(team_entered()));

	right_vbox->addSpacing(5);

	QLabel *label_task = new QLabel("&Task:");
	right_vbox->addWidget(label_task);

	edit_task = new QLineEdit();
	edit_task->setMaxLength(4);
	right_vbox->addWidget(edit_task);
	label_task->setBuddy(edit_task);
	QObject::connect(edit_task, SIGNAL(returnPressed()), this, SLOT(task_entered()));

	right_vbox->addSpacing(20);

	QPushButton *b1 = new QPushButton("&New");
	QObject::connect(b1, SIGNAL(clicked(bool)), this, SLOT(new_button()));
	right_vbox->addWidget(b1);

	QPushButton *b2 = new QPushButton("L&oad");
	QObject::connect(b2, SIGNAL(clicked(bool)), this, SLOT(load_button()));
	right_vbox->addWidget(b2);

	QPushButton *b3 = new QPushButton("&Save");
	QObject::connect(b3, SIGNAL(clicked(bool)), this, SLOT(save_button()));
	right_vbox->addWidget(b3);

	QPushButton *b4 = new QPushButton("&Run");
	QObject::connect(b4, SIGNAL(clicked(bool)), this, SLOT(run_button()));
	right_vbox->addWidget(b4);

	right_vbox->addStretch(1);

	main_hbox = new QHBoxLayout;
	main_hbox->addLayout(left_vbox, 1);
	main_hbox->addLayout(right_vbox);

	main = new GryfMain;
	main->app = this;
	main->setMinimumSize(QSize(1024, 768));
	main->setWindowTitle("Gryf");
	main->setLayout(main_hbox);

	activate_cell(0, 0);

	main->show();
}

void GryfApp::status(const QString s)
{
	if (status_line) {
		QString px = "[" + QString::number(cell_i) + "," + QString::number(cell_j) + "]  ";
		status_line->setText(px + s);
	}
}

void GryfApp::cancel_entry()
{
	if (entry != "") {
		entry = "";
		if (cell_i >= 0)
			cells[cell_i][cell_j]->render();
		status("Edit cancelled");
	}
}

void GryfApp::activate_cell(int i, int j)
{
	cancel_entry();
	if (cell_i >= 0) {
		GryfCell *cell = cells[cell_i][cell_j];
		cell->setPalette(palette_idle);
	}
	cell_i = i, cell_j = j;
	if (cell_i >= 0) {
		GryfCell *cell = cells[cell_i][cell_j];
		cell->setPalette(palette_active);
	}
	status("");
}

void GryfApp::move_rel(int di, int dj)
{
	if (cell_i < 0)
		activate_cell(0, 0);
	else {
		int i = cell_i + di, j = cell_j + dj;
		if (i >= 0 && i < NUM_ROWS && j >= 0 && j < NUM_COLS)
			activate_cell(i, j);
	}
}

void GryfApp::move_abs(int i, int j)
{
	if (cell_i < 0)
		activate_cell(0, 0);
	else
		activate_cell(i, j);
}

void GryfApp::set_dir(int i, int j)
{
	dir_i = i, dir_j = j;
	status("Direction set to [" + QString::number(i) + "," + QString::number(j) + "]");
}

void GryfApp::redraw_program()
{
	for (int i=0; i < NUM_ROWS; i++)
		for (int j=0; j < NUM_COLS; j++)
			cells[i][j]->render();
}

void GryfApp::type_key(QString s)
{
	insn ins;
	int res = ins.partial_parse((entry + s).toStdString());
	if (res <= 0)
		status("Unknown key: " + s);
	else if (res == 3) {
		pgm->ins[cell_i][cell_j] = ins;
		entry = "";
		cells[cell_i][cell_j]->render();
		move_rel(dir_i, dir_j);
	} else {
		status("Editing");
		entry += s;
		cells[cell_i][cell_j]->render_entry(&ins, res);
	}
}

bool GryfApp::grid_key_press(QKeyEvent *evt)
{
	switch (evt->key()) {
		case Qt::Key_Left:
			if (evt->modifiers() & Qt::ShiftModifier)
				set_dir(0, -1);
			else
				move_rel(0, -1);
			return true;
		case Qt::Key_Right:
			if (evt->modifiers() & Qt::ShiftModifier)
				set_dir(0, 1);
			else
				move_rel(0, 1);
			return true;
		case Qt::Key_Up:
			if (evt->modifiers() & Qt::ShiftModifier)
				set_dir(-1, 0);
			else
				move_rel(-1, 0);
			return true;
		case Qt::Key_Down:
			if (evt->modifiers() & Qt::ShiftModifier)
				set_dir(1, 0);
			else
				move_rel(1, 0);
			return true;
		case Qt::Key_PageUp:
			move_abs(0, cell_j);
			return true;
		case Qt::Key_PageDown:
			move_abs(NUM_ROWS-1, cell_j);
			return true;
		case Qt::Key_Home:
			move_abs(cell_i, 0);
			return true;
		case Qt::Key_End:
			move_abs(cell_i, NUM_COLS-1);
			return true;
		case Qt::Key_Return:
		case Qt::Key_Enter:
			move_rel(dir_i, dir_j);
			return true;
		case Qt::Key_Space:
			set_dir(0, 0);
			return true;
		case Qt::Key_Delete:
			cancel_entry();
			if (cell_i >= 0) {
				pgm->ins[cell_i][cell_j] = insn();
				cells[cell_i][cell_j]->render();
				move_rel(dir_i, dir_j);
			}
			return true;
		case Qt::Key_Insert:
			cancel_entry();
			if (cell_i >= 0) {
				if (evt->modifiers() & Qt::ControlModifier) {
					clipboard = pgm->ins[cell_i][cell_j];
				} else if (evt->modifiers() & Qt::ShiftModifier) {
					pgm->ins[cell_i][cell_j] = clipboard;
					cells[cell_i][cell_j]->render();
				} else {
					int i = cell_i, j = cell_j;
					move_rel(dir_i, dir_j);
					pgm->ins[i][j] = pgm->ins[cell_i][cell_j];
					cells[i][j]->render();
				}
			}
			return true;
		case Qt::Key_Escape:
			cancel_entry();
			return true;
		case Qt::Key_F1:
			cancel_entry();
			show_help();
			return true;
		default:
			QString s = evt->text();
			if (s != "") {
				type_key(s);
				return true;
			}
	}
	return false;
}

void GryfApp::team_entered()
{
	edit_task->setFocus(Qt::FocusReason::OtherFocusReason);
}

void GryfApp::task_entered()
{
	grid_widget->setFocus(Qt::FocusReason::OtherFocusReason);
}

void GryfApp::new_button()
{
	cancel_entry();

	if (QMessageBox::question(main, tr("Confirmation"), tr("Clear existing program?"), QMessageBox::Ok | QMessageBox::Cancel) != QMessageBox::Ok)
		return;

	debug << "Clearing program" << endl;
	program *p = new program;
	delete pgm;
	pgm = p;
	redraw_program();
}

void GryfApp::load_button()
{
	cancel_entry();

	QString name = QFileDialog::getOpenFileName(main, tr("Open file..."), "src");
	if (name == "")
		return;
	debug << "Loading from " << name << endl;

	program *p = new program;
	try {
		p->load(name.toStdString());
		debug << "Loaded" << endl;
		delete pgm;
		pgm = p;
		redraw_program();
	} catch (const hippo_error &e) {
		delete p;
		QString pos;
		if (e.i >= 0)
			pos = QString("Chyba na pozici [") + QString::number(e.i) + QString(",") + QString::number(e.j) + QString("]:\n");
		debug << "Load failed: " << pos << QString::fromStdString(e.what()) << endl;
		QMessageBox::critical(main, tr("Load error"), tr("Nepodařilo se načíst:\n") + pos + QString::fromStdString(e.what()), QMessageBox::Ok);
	}

	QString n = name;
	int slash = n.lastIndexOf('/');
	if (slash >= 0)
		n = n.mid(slash+1);
	int i = n.indexOf('-');
	if (i >= 0) {
		edit_team->setText(n.left(i));
		edit_task->setText(n.mid(i+1));
	}
}

void GryfApp::save_button()
{
	cancel_entry();

	const QString team = edit_team->text();
	const QString task = edit_task->text();

	mkdir("src", 0777);
	QString name = QFileDialog::getSaveFileName(main, tr("Save file..."), "src/" + team + "-" + task, QString(), Q_NULLPTR, QFileDialog::DontConfirmOverwrite);
	if (name == "")
		return;
	debug << "Saving to " << name << endl;
	pgm->save(name.toStdString());
	debug << "Saved" << endl;
}

void GryfApp::run_button()
{
	cancel_entry();

	const QString team = edit_team->text();
	const QString task = edit_task->text();

	if (team == "" || task == "") {
		QMessageBox::critical(main, tr("Nelze spustit"), tr("Je nutné zadat úlohu a tým"), QMessageBox::Ok);
		return;
	}

	try {
		pgm->resolve_labels();
		state st(pgm);
		const QString in = QString("task/") + task + QString(".in");
		st.read_input(in.toStdString());
		st.run();
		debug << "Stopped in " << st.run_time << " steps." << endl;

		mkdir("test", 0777);
		const QString out = QString("test/") + team + QString("-") + task + QString(".out");
		st.write_output(out.toStdString());

		QString correct_out = QString("task/") + task + QString(".out");
		QString judge = QString("task/") + task + QString(".judge");
		bool ok = false;
		if (QFileInfo(judge).exists()) {
			QString run_judge = judge + QString(" ") + in + QString(" ") + out + QString(" ") + correct_out;
			debug << "Running " << run_judge << endl;
			int rc = system(run_judge.toUtf8());
			debug << "\tExit code " << rc << endl;
			if (rc)
				QMessageBox::critical(main, tr("Chybný výstup"), tr("Chybný výstup\nJudge vrátil kód ") + QString::number(rc), QMessageBox::Ok);
			else
				ok = true;
		} else if (QFileInfo(correct_out).exists()) {
			QString run_diff = QString("diff -w >/dev/null ") + out + QString(" ") + correct_out;
			debug << "Running " << run_diff << endl;
			int rc = system(run_diff.toUtf8());
			debug << "\tExit code " << rc << endl;
			if (rc)
				QMessageBox::critical(main, tr("Chybný výstup"), tr("Chybný výstup\n(diff nesouhlasí)"), QMessageBox::Ok);
			else
				ok = true;
		} else {
			QMessageBox::critical(main, tr("Chybný výstup"), tr("Nevím, jak zkontrolovat výstup"), QMessageBox::Ok);
		}
		if (ok) {
			QMessageBox::information(main, tr("Správný výstup"), tr("Výstup je v pořádku"), QMessageBox::Ok);
			QString okay = QString("test/") + team + QString("-") + task + QString(".ok");
			fstream fo(okay.toStdString(), fo.out | fo.trunc);
			if (!fo.is_open())
				throw hippo_error("Nemohu vytvořit soubor s body!");
		}
	} catch (hippo_error &e) {
		QString pos;
		if (e.i >= 0)
			pos = QString("Chyba na pozici [") + QString::number(e.i) + QString(",") + QString::number(e.j) + QString("]:\n");
		QMessageBox::critical(main, tr("Běhová chyba"), pos + e.what(), QMessageBox::Ok);
		if (e.i >= 0)
			activate_cell(e.i, e.j);
	}
}

void GryfApp::show_help()
{
	QMessageBox::information(main, tr("Help!"), tr(
		"Ovládání:\n\n"
		"Šipky pohybují po programu\n"
		"Shift+šipky určují směr pohybu\n"
		"Mezera vypíná pohyb\n"
		"Home/End/PgUp/PgDn skáčí na okraje\n"
		"Delete maže instrukci\n"
		"Insert kopíruje následující instrukci ve směru\n"
		"Ctrl+Insert/Shift+Insert jsou cut/paste jedné instrukce\n"
		"Escape přepíná focus na program\n"
		"F1 zobrazuje help\n"
	), QMessageBox::Ok);
}

// GryfCell

void GryfCell::mouseReleaseEvent(QMouseEvent *evt)
{
	if (evt->button() == Qt::LeftButton)
		app->activate_cell(i, j);
}

static QString escape(QString &s)
{
	QString out = "";
	for (QChar c: s) {
		if (c == '<')
			out += "&lt;";
		else if (c == '>')
			out += "&gt;";
		else if (c == '&')
			out += "&amp;";
		else
			out += c;
	}
	return out;
}

void GryfCell::render()
{
	insn *ins = &app->pgm->ins[i][j];

	QString iname = QString::fromStdString(ins->def->show_pretty(ins));

	QString syn = "";
#if 0
	insn_def *def = ins->def;
	for (int i=0; def->syntax[i] && !(def->syntax[i] >= 'a' && def->syntax[i] <= 'z'); i++)
		syn += def->syntax[i];
#endif

	QString desc = "";
	if (iname != "") {
		desc += escape(iname);
		if (syn != "")
			desc += "  <b style='color: #cccccc'>(" + escape(syn) + ")</b>";
	}

	setText(desc);
}

void GryfCell::render_entry(insn *ins, int parse_result)
{
	if (parse_result < 2) {
		setText("<b>" + escape(app->entry) + "</b>");
		return;
	}

	QString iname = QString::fromStdString(ins->def->show_pretty(ins));
	QString desc = escape(iname);

	setText(desc);
}

// GryfGrid

void GryfGrid::keyPressEvent(QKeyEvent *evt)
{
	// debug << "GRID keypress: " << evt->key() << endl;
	if (!app->grid_key_press(evt))
		QWidget::keyPressEvent(evt);
}

// GryfMain

void GryfMain::keyPressEvent(QKeyEvent *evt)
{
	// debug << "MAIN keypress: " << evt->key() << endl;
	switch (evt->key()) {
		case Qt::Key_Escape:
			app->grid_widget->setFocus(Qt::FocusReason::OtherFocusReason);
			break;
		default:
			QWidget::keyPressEvent(evt);
	}
}

// main

int main(int argc, char **argv)
{
	app = new GryfApp(argc, argv);
	app->exec();
	return 0;
}
